import pandas as pd

from zgmmodule.utils.utils import search_file
from zgmmodule.utils.logger import logger


def get_rec_syn_df(query_g: str, target_g: str, work_dir: str) -> pd.DataFrame:
    q_t_synteny_fp, t_q_synteny_fp = search_file(
        work_dir, 'anchor', query_g, target_g, type='rec')
    logger.info('Start to combine syteny evidence')
    logger.info(
        f'Searching Files:\n query -> target synteny file: {q_t_synteny_fp}\n target -> query synteny file: {t_q_synteny_fp}')
    exclude_fwd = [i for i, line in enumerate(
        open(q_t_synteny_fp)) if line.startswith('#')]
    exclude_rev = [i for i, line in enumerate(
        open(t_q_synteny_fp)) if line.startswith('#')]

    fwd_df = pd.read_csv(q_t_synteny_fp, sep='\t', header=None,
                         usecols=[0, 1], skiprows=exclude_fwd)
    rev_df = pd.read_csv(t_q_synteny_fp, sep='\t', header=None,
                         usecols=[0, 1], skiprows=exclude_rev)
    fwd_df.columns = ['query_g', 'target_g']
    rev_df.columns = ['target_g', 'query_g']

    rev_syn_df = pd.merge(fwd_df,
                          rev_df,
                          on='query_g',
                          how='inner')
    rev_syn_df = rev_syn_df[rev_syn_df['target_g_x'] ==
                            rev_syn_df['target_g_y']].reset_index(drop=True)
    rev_syn_df.drop(columns=['target_g_y'], inplace=True)
    rev_syn_df.columns = ['genename', 'synteny']
    rev_syn_df = rev_syn_df.groupby('genename', as_index=False).agg(
        lambda x: ','.join(list(x)) if len(x) > 1 else x)
    return rev_syn_df



from dataclasses import dataclass, field
from typing import *
import sys
import os
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(BASE_DIR)
from bin.GeneMap import main


@dataclass
class TestArgs:
    list: str = 'tests/data/Zm-B73v5.genelist'
    dir: str = '.'
    query_g: str = 'Zm-B73v5'
    target_g: str = 'Zv-TEO01'
    output: str = 'tests/data/test'
    evidence_l: List = field(default_factory=lambda: [
                             'rbh', 'crossmap', 'synteny', 'ortholog'])
    ratio: float = 0.8
    overlap: float = 0.8


TestArgs = TestArgs()

main(TestArgs)



import pandas as pd
from zgmmodule.utils.utils import search_file
from zgmmodule.utils.logger import logger


def map_str_list(x: str) -> list:
    lst = x.split(',')
    lst = [i.strip() for i in lst]
    lst = [i.split(' ')[1] if len(i.split(' ')) == 2 else i for i in lst]
    if len(lst) == 1:
        return lst[0]
    return lst


def get_ortholog_df(query_g: str, target_g: str, work_dir: str):
    """
    get ortholog df
    """
    ortholog_fp = search_file(work_dir, 'tsv', type='ortholog')
    logger.info('Start to combine ortholog evidence')
    logger.info(f"Searching Files:\n ortholog file: {ortholog_fp}")
    used_cols = [query_g, target_g]
    ortholog_df = pd.read_csv(ortholog_fp, sep='\t', header=0,
                              usecols=used_cols)[used_cols]
    ortholog_df.fillna('None', inplace=True)
    ortholog_df[query_g] = ortholog_df[query_g].map(map_str_list)
    ortholog_df[target_g] = ortholog_df[target_g].map(map_str_list)
    ortholog_df = ortholog_df.explode(query_g)
    ortholog_df = ortholog_df.reset_index(drop=True)
    ortholog_df.columns = ['genename', 'ortholog']
    ortholog_df['ortholog'] = ortholog_df['ortholog'].map(
        lambda x: ','.join(x) if isinstance(x, list) else x)
    # ortholog_df = ortholog_df.groupby('genename', as_index=False).agg(
    #     lambda x: list(x) if len(x) > 1 else x)

    return ortholog_df



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@文件        :GeneMap.py
@说明        :
@时间        :2022/07/18 21:02:56
@作者        :wjwei
@版本        :0.01
@邮箱        :wjwei9908@gmail.com
'''

import argparse

import sys
from typing import *
import os
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(BASE_DIR + '/src/')
from zgmmodule.utils import utils
from zgmmodule.utils.logger import pretty_print


LEGAL_EVIDENCES = ['rbh', 'crossmap', 'synteny', 'ortholog']

LOGO = """[bold green]
 _____                ___  ___            |
|  __ \               |  \/  |            | Author: Wenjie Wei
| |  \/ ___ _ __   ___| .  . | __ _ _ __  | Email: wjwei9908@gmail.com
| | __ / _ \ '_ \ / _ \ |\/| |/ _` | '_ \ | Contribute: Wenjie Wei;Songtao Gui;Han Xiao
| |_\ \  __/ | | |  __/ |  | | (_| | |_) || Department: HZAU-Maize-Bioinformatics
 \____/\___|_| |_|\___\_|  |_/\__,_| .__/ |
                                   | |    |
                                   |_|    |
demo: GeneMap.py -l list.txt -d work_dir -q GenomeA -t GenomeB -o output_prefix
"""


def get_args():
    # define arguments
    parser = argparse.ArgumentParser(
        description=None, prog='GeneMap.py',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # required arguments
    required_parser = parser.add_argument_group('required arguments')
    required_parser.add_argument('-l', '--list', action='store',
                                 help='genename list file', required=True, dest='list')
    required_parser.add_argument('-d', '--dir', action='store',
                                 help='work dir', required=True, default='.', dest='dir')
    required_parser.add_argument('-q', '--query', action='store',
                                 help='query genome', required=True, dest='query_g')
    required_parser.add_argument('-t', '--target', action='store',
                                 help='target genome', required=True, dest='target_g')
    required_parser.add_argument('-o', '--output', action='store',
                                 help='output prefix', required=True, dest='output')

    # options arguments
    option_parser = parser.add_argument_group('optional arguments')
    option_parser.add_argument('-e', '--evidence',
                               help="evidence list", nargs='+', dest='evidence_l',
                               default=['rbh', 'crossmap', 'synteny', 'ortholog'])
    option_parser.add_argument('-r', '--ratio', action='store',
                               help='crossmap ratio', default=0.8, dest='ratio')
    option_parser.add_argument('-p', '--overlap', action='store',
                               help='overlap fraction', default=0.8, dest='overlap')

    return parser.parse_args()


def main(args):
    # judge the evidence
    utils.check_in_list(args.evidence_l, LEGAL_EVIDENCES,
                        print_list=True)

    # get raw genelist df
    raw_df = utils.get_raw_df(args.list)

    # add evidence and output raw df
    raw_evd_df = utils.get_raw_evd_df(args, raw_df, write=True)

    # process raw result
    utils.process_raw(args, raw_evd_df)


# TODO: README 4) 学会单元测试 5) 学会setup 6)发布

if __name__ == '__main__':
    pretty_print(LOGO)
    # get args
    args = get_args()
    main(args)



import sys
import pandas as pd
import pybedtools
from zgmmodule.chain import chain

from zgmmodule.utils.utils import search_file
from zgmmodule.utils.logger import logger


NULL_REGION = 'chrnull\t0\t0\t+'


def get_region(x, min_ratio: float = 0.8):
    chrom = x['chrom']
    start = x['start']
    end = x['end']
    strand = x['strand']
    # genename = x['genename']
    matches = chain.map_coordinates(bx_ivl, chrom, start, end, strand)
    if (matches is None) or (len(matches) % 2 != 0):
        return NULL_REGION
    # when matches == 2, there is one-to-one match (i.e. 100% match)
    if len(matches) == 2:
        region = '\t'.join([str(_) for _ in matches[1]])
        return region
    if len(matches) > 2:
        query_m = matches[::2]
        query_m_nt = sum([i[2] - i[1] for i in query_m])  # sum([3,2])
        target_m = matches[1::2]
        target_m_chroms = set([i[0] for i in target_m])
        target_m_starts = [i[1] for i in target_m]
        target_m_ends = [i[2] for i in target_m]
        target_m_strand = set([i[3] for i in target_m]).pop()
        map_ratio = query_m_nt / (end - start)
        if map_ratio >= min_ratio:
            if len(target_m_chroms) == 1:
                target_m_chrom = target_m_chroms.pop()
                target_m_start = min(target_m_starts)
                target_m_end = max(target_m_ends)
                region = '\t'.join([target_m_chrom, str(
                    target_m_start), str(target_m_end), target_m_strand])
                return region
            else:
                return NULL_REGION
        else:
            return NULL_REGION


# def string_bed_ivl(x, min_ovp: float = 0.8):
#     if x != 'None':
#         str_bed = pybedtools.bedtool.BedTool(x, from_string=True)
#         intersect_ivl = target_bed.intersect(
#             str_bed, nonamecheck=True, F=min_ovp, sorted=True)
#         if intersect_ivl:
#             result = []
#             intersect_ivl_lines = intersect_ivl.__str__().splitlines()
#             for line in intersect_ivl_lines:
#                 line_chr, line_start, line_end, line_junk, line_name, line_strand = line.split(
#                     '\t')
#                 result.append(
#                     f'{line_name}@({line_chr}:{line_start}-{line_end}-{line_strand})')
#             str_result = ','.join(result)
#             return str_result
#         else:
#             result = []
#             for ivl in str_bed:
#                 result.append(
#                     f'({ivl.chrom}:{ivl.start}-{ivl.end})')
#             str_result = ','.join(result)
#             return str_result
#     else:
#         return 'None'

def get_intersect(region_df: pd.DataFrame, overlap: float = 0.8) -> pybedtools.bedtool.BedTool:
    """
    Get intersect between target and query bed files.
    """
    region_list = region_df.to_list()
    region_bed_str = '\n'.join(region_list)
    region_bed = pybedtools.bedtool.BedTool(region_bed_str, from_string=True)
    intersect_bed = region_bed.intersect(target_bed, nonamecheck=True,
                                         f=overlap, wo=True, loj=True)
    return intersect_bed


def get_intersect_df(intersect_bed: pybedtools.bedtool.BedTool) -> pd.DataFrame:
    """_summary: Get intersect df from a bed.

    Args:
        intersect_bed (pybedtools.bedtool.BedTool): _description_

    Returns:
        pd.DataFrame: _description_
    """
    lst = []
    for ivl in intersect_bed:
        chrom = ivl.chrom
        region_start, region_end = ivl.start, ivl.end
        gene_start, gene_end = ivl.fields[5:7]
        if chrom == 'chrnull':
            lst.append('NoneRegion')
        else:
            if gene_start == '-1' or gene_end == '-1':
                lst.append(f'{chrom}:{region_start}-{region_end}')
            else:
                target_gene = ivl.fields[7]
                lst.append(
                    f'{target_gene}@{chrom}:{max([region_start,int(gene_start)])}-{min([region_end,int(gene_end)])}')
    df = pd.DataFrame(lst, columns=['crossmap'])
    return df


def get_crossmap_df(raw_df: pd.DataFrame,
                    query_g: str,
                    target_g: str,
                    work_dir: str,
                    ratio: float = 0.8,
                    overlap: float = 0.8):
    chain_fp = search_file(work_dir, 'chain.gz',
                           query_g=query_g, target_g=target_g, type='single')
    logger.info('Start to combine crossmap evidence')
    logger.info(f"Searching Files:\n chain file: {chain_fp}")
    logger.info(f"Reading chain file: {chain_fp}")
    global bx_ivl
    bx_ivl = chain.read_chain_file(chain_fp)[0]
    q_bed_fp = search_file(work_dir, 'bed', single_g=query_g, type='single')
    logger.info(f"Searching Files:\n query genome bed file:{q_bed_fp}")
    t_bed_fp = search_file(work_dir, 'bed', single_g=target_g, type='single')
    logger.info(f"Searching Files:\n taget genome bed file: {t_bed_fp}")
    q_bed_df = pd.read_csv(q_bed_fp, sep='\t', header=None)
    q_bed_df.columns = ['chrom', 'start', 'end', 'genename', 'junk', 'strand']
    raw_bed_df = pd.merge(raw_df, q_bed_df, on='genename',
                          how='left').drop(columns=['junk'])
    global target_bed
    target_bed = pybedtools.BedTool(t_bed_fp).sort()
    raw_bed_df['region'] = raw_bed_df.apply(
        get_region, axis=1, args=(ratio,))
    intersect_bed = get_intersect(raw_bed_df['region'], overlap)
    intersect_df = get_intersect_df(intersect_bed)

    # concat method
    if len(intersect_df) == len(raw_bed_df):
        return pd.concat([raw_bed_df['genename'], intersect_df], axis=1)
    else:
        logger.error('intersect_df and raw_bed_df have different length')
        sys.exit(1)

    # raw_bed_df['crossmap'] = raw_bed_df['region'].progress_apply(
    #     string_bed_ivl, args=(overlap,))
    # return raw_bed_df[['genename', 'crossmap']]



from subprocess import Popen
import urllib
import gzip
import bz2
import sys
import os
import glob
import time
import pandas as pd
from tqdm import tqdm
from zgmmodule.utils.logger import logger, console

tqdm.pandas(desc='Processing', colour='green')


def _warn(msg):
    logger.warning(f"{msg}")


def _error(msg):
    logger.error(f"{msg}")
    sys.exit(1)


def _info(msg):
    logger.info(f"{msg}")


def search_file(work_dir: str,
                surfix: str,
                query_g: str = '', target_g: str = '',
                type: str = 'rec',
                single_g: str = ''):
    abs_dir_path = os.path.abspath(work_dir)
    if type == 'rec':
        forward_fp = f"{query_g}.{target_g}.{surfix}"
        reverse_fp = f"{target_g}.{query_g}.{surfix}"
        fw_search = glob.glob(
            f"{abs_dir_path}/**/{forward_fp}", recursive=True)
        rv_search = glob.glob(
            f"{abs_dir_path}/**/{reverse_fp}", recursive=True)
        if len(fw_search) == 1 and len(rv_search) == 1:
            return fw_search[0], rv_search[0]
        else:
            _error(
                f"{forward_fp} or {reverse_fp} not found or duplicated in your work dir : {abs_dir_path}")
    elif type == 'single':
        if query_g and target_g:
            chain_fp = f"{query_g}.{target_g}.{surfix}"
            chain_search = glob.glob(
                f"{abs_dir_path}/**/{chain_fp}", recursive=True)
            if len(chain_search) == 1:
                return chain_search[0]
            else:
                _error(
                    f"{chain_fp} not found or duplicated in your work dir : {abs_dir_path}")
        single_fp = f"{single_g}.{surfix}"
        single_search = glob.glob(
            f"{abs_dir_path}/**/{single_fp}", recursive=True)
        if len(single_search) == 1:
            return single_search[0]
        else:
            _error(
                f"{single_fp} not found or duplicated in your work dir : {abs_dir_path}")
    elif type == 'ortholog':
        ortholog_fp = f"rthogroups.{surfix}"
        ortholog_search = glob.glob(
            f"{abs_dir_path}/**/?{ortholog_fp}", recursive=True)
        if len(ortholog_search) == 1:
            return ortholog_search[0]
        else:
            _error(
                f"O[o]{ortholog_fp} not found or duplicated in your work dir : {abs_dir_path}")
    else:
        _error('type error, contact the author')


def nopen(f, mode="rb"):
    if not isinstance(f, str):
        return f
    if f.startswith("|"):
        p = Popen(f[1:], stdout=PIPE, stdin=PIPE, shell=True)
        if mode[0] == "r":
            return p.stdout
        return p
    return {"r": sys.stdin, "w": sys.stdout}[mode[0]] if f == "-" \
        else gzip.open(f, mode) if f.endswith((".gz", ".Z", ".z")) \
        else bz2.BZ2File(f, mode) if f.endswith((".bz", ".bz2", ".bzip2")) \
        else urllib.urlopen(f) if f.startswith(("http://", "https://", "ftp://")) \
        else open(f, mode)


def reader(fname):
    for l in nopen(fname):
        yield l.decode('utf8').strip().replace("\r", "")


def check_in_list(choice_list, legal_list, print_list: bool = True):
    if print_list:
        _info(f"choiced evidence: {choice_list}")
    if any(item not in legal_list for item in choice_list):
        _error(f'evidence error: please select in {legal_list}')


def get_time(func):
    """ a decorator to get the time of a function and return it """
    def wrapper(*args, **kwargs):
        start = time.time()
        res = func(*args, **kwargs)
        end = time.time()
        _info(f"Combine {func.__name__} evidenve cost {end - start}s ✅")
        return res
    return wrapper


def get_raw_df(genelist_fp: str) -> pd.DataFrame:
    # read input gene list
    try:
        raw_df = pd.read_csv(genelist_fp, sep='\t', header=None)
        _info(f"Read input genelist: {genelist_fp} ✅")
    except Exception as e:
        _error(f"Read {genelist_fp} failed, please check the file")
    if not raw_df.shape[1] == 1:
        _error('genelist file format error, except one column only')
    raw_df.columns = ['genename']
    return raw_df


@get_time
def rbh(args, raw_df):
    from zgmmodule.blast import blast
    rbh_df = blast.get_rbh_df(args.query_g, args.target_g, args.dir)
    raw_rbh_merge = pd.merge(
        raw_df, rbh_df, on='genename', how='left')
    return raw_rbh_merge


@get_time
def ortholog(args, raw_df):
    from zgmmodule.ortholog import ortholog
    ortholog_df = ortholog.get_ortholog_df(
        args.query_g, args.target_g, args.dir)
    raw_ortholog_merge = pd.merge(
        raw_df, ortholog_df, on='genename', how='left')
    if len(raw_ortholog_merge) != len(raw_df):
        _error("Ortholog evidence error")
    return raw_ortholog_merge


@get_time
def synteny(args, raw_df):
    from zgmmodule.synteny import synteny
    rec_syn_df = synteny.get_rec_syn_df(
        args.query_g, args.target_g, args.dir)
    raw_syn_merge = pd.merge(
        raw_df, rec_syn_df, on='genename', how='left')
    return raw_syn_merge


@get_time
def crossmap(args, raw_df):
    from zgmmodule.chain import crossmap
    crossmap_df = crossmap.get_crossmap_df(
        raw_df, args.query_g,
        args.target_g, args.dir,
        ratio=args.ratio, overlap=args.overlap
    )
    raw_crossmap_merge = pd.merge(
        raw_df, crossmap_df, on='genename', how='left')
    return raw_crossmap_merge


def get_raw_evd_df(args, raw_df, write: bool = True) -> pd.DataFrame:
    # get raw df
    # 1 rbh
    if 'rbh' in args.evidence_l:
        output_df = rbh(args, raw_df)
    else:
        output_df = raw_df
        _warn('rbh not in evidence list')
    # 2 ortholog
    if 'ortholog' in args.evidence_l:
        output_df = ortholog(args, output_df)
    else:
        _warn('ortholog not in evidence list')
    # 3 synteny
    if 'synteny' in args.evidence_l:
        output_df = synteny(args, output_df)
    else:
        _warn('synteny not in evidence list')
    # 4 crossmap
    if 'crossmap' in args.evidence_l:
        output_df = crossmap(args, output_df)
    else:
        _warn('crossmap not in evidence list')
    output_df.fillna('None', inplace=True)
    if write:
        _info(f'Writing raw result to {args.output}_raw.tsv')
        output_df.to_csv(
            f"{args.output}_raw.tsv", sep='\t', index=False, header=True)
    return output_df


def map_to_list(x):
    if '@' in x:
        lst = []
        for i in x.split(','):
            g = x.split('@')[0]
            if g not in lst:
                lst.append(g)
        return lst
    if x == 'NoneRegion':
        return ['None']
    else:
        return [item.strip() for item in x.split(',')]


def process_raw(args, raw_evd_df: pd.DataFrame) -> None:
    with console.status(f"[bold green]Processing raw result to final result") as s:
        process_df = raw_evd_df.melt(
            id_vars=['genename'], var_name='evd', value_name='target')
        process_df['target'] = process_df['target'].map(map_to_list)
        process_df = process_df.explode('target')
        process_df = process_df.drop(
            process_df[process_df.target == 'None'].index).reset_index(drop=True)
        # process_df = process_df.drop(
        #     process_df[process_df.target == 'NoneRegion'].index).reset_index(drop=True)
        result_df = process_df.groupby(['genename', 'target'], as_index=False).agg({
            'evd': lambda x: ','.join(x)})
        result_df.to_csv(
            f"{args.output}_result.tsv", sep='\t', index=False, header=True)
    _info(f'Writing final result to {args.output}_final.tsv')



import logging
from rich.logging import RichHandler
from rich import print
from rich.console import Console

console = Console()
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    handlers=[RichHandler(rich_tracebacks=True)])
logger = logging.getLogger("GeneMap")
pretty_print = print.__call__



from zgmmodule.utils import utils
from bx.intervals.intersection import Interval, Intersecter
from rich.progress import track


def read_chain_file(chain_file, print_table=False):
    '''
    Read chain file.

    Parameters
    ----------
    chain_file : file
            Chain format file. Input chain_file could be either plain text, compressed file
            (".gz",".Z", ".z", ".bz", ".bz2", ".bzip2"), or a URL pointing to the chain file
            ("http://","https://", "ftp://"). If url was used, chain file must be plain text.

    print_table : bool, optional
            Print mappings in human readable table.

    Returns
    -------
    maps : dict
            Dictionary with source chrom name as key, IntervalTree object as value. An
            IntervalTree contains many intervals. An interval is a start and end position
            and a value. eg. Interval(11, 12, strand="-", value = "abc")

    target_chromSize : dict
            Chromosome sizes of target genome

    source_chromSize : dict
            Chromosome sizes of source genome
    '''

    maps = {}
    target_chromSize = {}
    source_chromSize = {}
    if print_table:
        blocks = []
    total_size = len(list(utils.reader(chain_file)))
    for line in track(utils.reader(chain_file), description="[bold green]Reading", total=total_size,):
        # Example: chain 4900 chrY 58368225 + 25985403 25985638 chr5 151006098 - 43257292 43257528 1
        if not line.strip():
            continue
        line = line.strip()
        if line.startswith(('#', ' ')):
            continue
        fields = line.split()

        if fields[0] == 'chain' and len(fields) in [12, 13]:
            # score = int(fields[1])		  # Alignment score
            source_name = fields[2]		  # E.g. chrY
            source_size = int(fields[3])  # Full length of the chromosome
            source_strand = fields[4]	  # Must be +
            if source_strand != '+':
                raise Exception(
                    "Source strand in a chain file must be +. (%s)" % line)
            source_start = int(fields[5])  # Start of source region
            # source_end = int(fields[6])	  # End of source region

            target_name = fields[7]		  # E.g. chr5
            target_size = int(fields[8])  # Full length of the chromosome
            target_strand = fields[9]	  # + or -
            target_start = int(fields[10])
            # target_end = int(fields[11])
            target_chromSize[target_name] = target_size
            source_chromSize[source_name] = source_size

            if target_strand not in ['+', '-']:
                raise Exception("Target strand must be - or +. (%s)" % line)
            # chain_id = None if len(fields) == 12 else fields[12]
            if source_name not in maps:
                maps[source_name] = Intersecter()

            sfrom, tfrom = source_start, target_start

        # Now read the alignment chain from the file and store it as a list (source_from, source_to) -> (target_from, target_to)
        elif fields[0] != 'chain' and len(fields) == 3:
            size, sgap, tgap = int(fields[0]), int(fields[1]), int(fields[2])
            if print_table:
                if target_strand == '+':
                    blocks.append((source_name, sfrom, sfrom + size, source_strand,
                                  target_name, tfrom, tfrom + size, target_strand))
                elif target_strand == '-':
                    blocks.append((source_name, sfrom, sfrom + size, source_strand, target_name,
                                  target_size - (tfrom + size), target_size - tfrom, target_strand))

            if target_strand == '+':
                maps[source_name].add_interval(
                    Interval(sfrom, sfrom + size, (target_name, tfrom, tfrom + size, target_strand)))
            elif target_strand == '-':
                maps[source_name].add_interval(Interval(
                    sfrom, sfrom + size, (target_name, target_size - (tfrom + size), target_size - tfrom, target_strand)))

            sfrom += size + sgap
            tfrom += size + tgap

        elif fields[0] != 'chain' and len(fields) == 1:
            size = int(fields[0])
            if print_table:
                if target_strand == '+':
                    blocks.append((source_name, sfrom, sfrom + size, source_strand,
                                  target_name, tfrom, tfrom + size, target_strand))
                elif target_strand == '-':
                    blocks.append((source_name, sfrom, sfrom + size, source_strand, target_name,
                                  target_size - (tfrom + size), target_size - tfrom, target_strand))

            if target_strand == '+':
                maps[source_name].add_interval(
                    Interval(sfrom, sfrom + size, (target_name, tfrom, tfrom + size, target_strand)))
            elif target_strand == '-':
                maps[source_name].add_interval(Interval(
                    sfrom, sfrom + size, (target_name, target_size - (tfrom + size), target_size - tfrom, target_strand)))
        else:
            raise Exception("Invalid chain format. (%s)" % line)
    # if (sfrom + size) != source_end  or (tfrom + size) != target_end:
    #	 raise Exception("Alignment blocks do not match specified block sizes. (%s)" % header)

    if print_table:
        for i in blocks:
            print('\t'.join([str(n) for n in i]))

    return (maps, target_chromSize, source_chromSize)


def intersectBed(lst1, lst2):
    '''
    Return intersection of two bed regions.

    Parameters
    ----------
    lst1 : list
            The 1st genomic region. List of chrom, start, end.
            Example: ['chr1',10, 100]

    lst2 : list
             The 2nd genomic region. List of chrom, start, end.
             Example: ['chr1',50, 120]

    Examples
    --------
    >>> intersectBed(['chr1',10, 100],['chr1',50, 120])
    ('chr1', 50, 100)
    >>> intersectBed(['chr1',10, 100],['chr1',20, 30])
    ('chr1', 20, 30)

    '''
    (chr1, st1, end1) = lst1
    (chr2, st2, end2) = lst2
    if int(st1) > int(end1) or int(st2) > int(end2):
        raise Exception("Start cannot be larger than end")
    if chr1 != chr2:
        return None
    if int(st1) > int(end2) or int(end1) < int(st2):
        return None
    return (chr1, max(st1, st2), min(end1, end2))


def map_coordinates(mapping, q_chr, q_start, q_end, q_strand='+'):
    '''
    Map coordinates from source (i.e. original) assembly to target (i.e. new) assembly.

    Parameters
    ----------
    mapping : dict
            Dictionary with source chrom name as key, IntervalTree object as value.

    q_chr : str
            Chromosome ID of query interval

    q_start : int
            Start position of query interval.

    q_end : int
            End position of query interval.

    q_strand : str
            Strand of query interval.

    '''

    matches = []
    complement = {'+': '-', '-': '+'}

    if q_chr in mapping:
        targets = mapping[q_chr].find(q_start, q_end)
    elif q_chr.replace('chr', '') in mapping:
        targets = mapping[q_chr.replace('chr', '')].find(q_start, q_end)
    elif ('chr' + q_chr) in mapping:
        targets = mapping['chr' + q_chr].find(q_start, q_end)
    else:
        return None
    if len(targets) == 0:
        return None
    elif len(targets) == 1:
        s_start = targets[0].start
        s_end = targets[0].end
        t_chrom = targets[0].value[0]
        # t_chrom = update_chromID(q_chr, t_chrom, chr_style = chrom_style)
        t_start = targets[0].value[1]
        t_end = targets[0].value[2]
        t_strand = targets[0].value[3]

        (chr, real_start, real_end) = intersectBed(
            (q_chr, q_start, q_end), (q_chr, s_start, s_end))
        l_offset = abs(real_start - s_start)
        # r_offset = real_end - s_end
        size = abs(real_end - real_start)

        matches.append((chr, real_start, real_end, q_strand))
        if t_strand == '+':
            i_start = t_start + l_offset
            if q_strand == '+':
                matches.append((t_chrom, i_start, i_start + size, t_strand))
            else:
                matches.append((t_chrom, i_start, i_start +
                               size, complement[t_strand]))
        elif t_strand == '-':
            i_start = t_end - l_offset - size
            if q_strand == '+':
                matches.append((t_chrom, i_start, i_start + size, t_strand))
            else:
                matches.append((t_chrom, i_start, i_start +
                               size, complement[t_strand]))
        else:
            raise Exception(
                "Unknown strand: %s. Can only be '+' or '-'." % q_strand)

    elif len(targets) > 1:
        for t in targets:
            s_start = t.start
            s_end = t.end
            t_chrom = t.value[0]
            # t_chrom = update_chromID(q_chr, t_chrom, chr_style=chrom_style)
            t_start = t.value[1]
            t_end = t.value[2]
            t_strand = t.value[3]

            (chr, real_start, real_end) = intersectBed(
                (q_chr, q_start, q_end), (q_chr, s_start, s_end))

            l_offset = abs(real_start - s_start)
            # r_offset = abs(real_end - s_end)
            size = abs(real_end - real_start)
            matches.append((chr, real_start, real_end, q_strand))
            if t_strand == '+':
                i_start = t_start + l_offset
                if q_strand == '+':
                    matches.append(
                        (t_chrom, i_start, i_start + size, t_strand))
                else:
                    matches.append(
                        (t_chrom, i_start, i_start + size, complement[t_strand]))
            elif t_strand == '-':
                i_start = t_end - l_offset - size
                if q_strand == '+':
                    matches.append(
                        (t_chrom, i_start, i_start + size, t_strand))
                else:
                    matches.append(
                        (t_chrom, i_start, i_start + size, complement[t_strand]))
            else:
                raise Exception(
                    "Unknown strand: %s. Can only be '+' or '-'." % q_strand)

    # if print_match:
    #     print(matches)
        # input: 'chr1',246974830,247024835
        # output: [('chr1', 246974830, 246974833, '+' ), ('chr1', 248908207, 248908210, '+' ), ('chr1', 247024833, 247024835, '+'), ('chr1', 249058210, 249058212,'+')]
        # [('chr1', 246974830, 246974833), ('chr1', 248908207, 248908210)]

    return matches



import pandas as pd
from zgmmodule.utils.logger import logger
from zgmmodule.utils.utils import search_file


# def get_rbh_result(gene_name, rbh_df):
#     result = []
#     filter_df = rbh_df.query(f'qseqid_x=="{gene_name}"')
#     if filter_df.empty:
#         return 'None'
#     else:
#         for index, row in filter_df.iterrows():
#             result.append(row['qseqid_y'])
#         result_str = ','.join(result)
#         return result_str


def get_rbh_df(query_g: str, target_g: str, work_dir: str, e_filter: float = 1e-10) -> pd.DataFrame:
    q_t_blast_fp, t_q_blast_fp = search_file(
        work_dir, 'blast', query_g, target_g, type='rec')
    logger.info('Start to combine rbh evidence')
    logger.info(
        f'Searching Files:\n query -> target blast file: {q_t_blast_fp}\n target -> query blast file: {t_q_blast_fp}')
    fwd_df = pd.read_csv(q_t_blast_fp, sep='\t', header=None)
    rev_df = pd.read_csv(t_q_blast_fp, sep='\t', header=None)
    column_names = ['qseqid', 'sseqid', 'pident', 'length', 'mismatch',
                    'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore']
    fwd_df.columns = column_names
    rev_df.columns = column_names
    fwd_df = fwd_df.drop_duplicates(subset=['qseqid'], keep='first')
    rev_df = rev_df.drop_duplicates(subset=['qseqid'], keep='first')
    rbh = pd.merge(fwd_df,
                   rev_df[['qseqid', 'sseqid']],
                   left_on='sseqid',
                   right_on='qseqid',
                   how='outer')
    rbh = rbh.loc[rbh.qseqid_x == rbh.sseqid_y]
    rbh = rbh.groupby(['qseqid_x', 'sseqid_x']).max().reset_index()[
        ['qseqid_x', 'sseqid_x']]
    rbh.columns = ['genename', 'rbh']
    return rbh
